# -*- coding: utf-8 -*-
### genetische-Algorithmen.py
# zum konkreten Beispiel hier siehe in der Funktion bewerte

# Mutationswahrscheinlichkeit und Selektionswahrscheinlichkeit
# sind positive ganze Zahlen zwischen 0 und 100 ( also % )

### ===== Entwurfsentscheidungen ====================================
# Auch wenn der Name 'genetischer-Algorithmus' anderes andeutet, es
# gibt nicht den genetischen Algorithmus, dies ist >>>einer<<< ! 
# Zu den einzelnen Entwurfsentscheidungen siehe auch bei den
# einzelnen Teilprozessen.
# In der Funktion selbst steckt z.B. die Entwurfsentscheidung, den
# Abbruch allein von der Zahl der Schritte abhängig zu machen. Man
# kann auch andere Abbruchkriterien wählen.

# Eine weitere wichtige Entwurfsentscheidung betrifft die Bewertung
# der Individuen in den Populationen. Wenn auch der eigentliche Sinn
# der GA für die Informatik in der Tatsache steckt, dass die Bewer-
# tungen erheblich 'billiger' sind als das Generieren von Lösungen,
# steckt in der Verwaltung dieser Werte ein abzuwägender Verwaltungs-
# aufwand. Ich habe mich hier daher so entschieden, während der
# Bearbeitung die Bewertungen an die Individuen zu koppeln.
# Jedes von ihnen stellt während der Bearbeitung daher ein Paar 
# (<Individuum>, <Bewertung>) dar!
# Das setzt voraus, dass beim jedem Erzeugen eines neuen Individuums
# sofort eine Bewertung durchgeführt wird.

import random

### ===== Zufallsgenerator-setzen ===================================
# Setzt den Zufallsgenerator
# Das fuehrt bei erneutem Aufruf zur selben Folge
#random.seed(100)

### ===== erzeuge-Individuum ========================================
# Erzeugt ein Individuum der Population.
# Dieser eigentliche Erzeugungsprozess ist problemabhängig.
# Fuer Testzwecke sind hier fuenfelementige Listen von Zufallszahlen
# gewählt.

def erzeugeIndividuum(anzahl=5):
    return [random.randint(1,100) for i in range(anzahl)]
# Hinweis: Technik heisst "list comprehension"

###Test:
##print(erzeugeIndividuum())

### ===== bewerte ===================================================
# Die Funktion ist problemabhaengig.
# Im Beispiel wird schlicht der Mittelwert der Zahlen ohne die
# kleinste und die groesste im Individuum berechnet und dann der
# Gesamtabstand der beiden von diesem Mittelwert verwendet.
# Eine Umkehrung der Wirkung lässt sich beispielsweise durch die
# auskommentierte Zeile erreichen.
def bewerte(individuum):
    groesste = max(individuum)
    kleinste = min(individuum)
    mittlere=[]+individuum ## echte Kopie
    mittlere.remove(groesste)
    mittlere.remove(kleinste)
    mw = int(sum(mittlere)/3)
    summe = abs(mw-groesste)*2 + abs(mw-kleinste)*2
    ##summe=int(10000/summe)
    return summe

###Test:
##print(bewerte([15, 20, 92, 8, 41]))

### ===== erzeuge-Population ========================================
# Erzeugt die Grundpopulation für den genetischen Prozess.
# Parameter sind: Anzahl der Individuen,
# Wert: Eine Liste von Individuen
def erzeugePopulation(anzahl=10):
    population = []
    for i in range(anzahl):
        population.append(erzeugeIndividuum())
    return population

###Test:
##print(erzeugePopulation())

### ===== aktionP ===================================================
# Hilfsfunktion (Praedikat), die entscheidet, ob eine Aktion stattfindet
# oder nicht. Die Entscheidung hängt von der übergebenen ganzzahligen
# Wahrscheinlichkeit p ab. 0 < p < 100
def aktionP(p):
    return (random.randint(1,100) <= p)

###Test:
##for i in range(10): print(aktionP(30))

### ===== waehle-Positionen =========================================
# Hilfsfunktion
# Wählt zwei verschiedene Zahlen aus {0 .. n-1} fuer die Position des
# Schnitts in den "Genen" aus und gib sie als Liste  zurück. 
# Die Zahlen sind stets geordnet: [klein gross]
# Unzulässige Parameter für n lösen einen Fehler aus!
def waehlePositionen(n):
    if type(n)!=int: return [0,n-1]
    if n<2: return[0,0]
    z1=random.randint(0,n-1)
    z2=random.randint(0,n-1)
    if (0 in[z1,z2]) and (n-1 in[z1,z2]):
        return waehlePositionen(n) # keine Auswahl, noch einmal versuchen
    if z1<z2: return [z1,z2]
    if z2<z1: return [z2,z1]
    return waehlePositionen(n) # gleich, noch einmal versuchen

###Test:
##print(waehlePositionen(30))

### ===== mutiere ===================================================
# Mutiert ein einzelnes Individuum.
# Hier muss die Bewertung bei den Individuen fehlen !
# Dieser eigentliche Mutationsprozess ist problemabhängig.
# Individuum -> Individuum

## Beispielversion:
def mutiere(individuum):
    index=random.randint(1,len(individuum)-1)
    wert=random.randint(1,100)
    if index==0: [wert]+individuum[1:]
    if index==len(individuum)-1: individuum[0:index]+[wert]
    return individuum[0:index]+[wert]+individuum[index+1:]

###Test:
##for i in range(10):
##    print(mutiere([15, 20, 92, 8, 41]))

### ===== Mutation ==================================================
# Steuert die Mutation für die gesamte Population.
# Die bei der Mutation neu erzeugten Individuen werden der Population
# hinzugefügt, dadurch ändert sich die Gesamtzahl. Deren Anpassung
# erfolgt bei der Selektion.
# Parameter sind: Population, Mutationswahrscheinlichkeit (pMutatioon)
# Wert:  Population zuzüglich der Mutierten
def mutation(population, pMutatioon):
    mutierte = []
    for individuum in population:
        if aktionP(pMutatioon):
            mutierte.append(mutiere(individuum))
        else:
            mutierte.append(individuum)
    return population+mutierte

###Test:
##population = erzeugePopulation()
##print(population)
##print()
##population = mutation(population, 40)
##for i in range(len(population)):
##    print(population[i])

### ===== kreuze ====================================================
# Kreuzt zwei einzelne Individuen.
# Hier muss die Bewertung bei den Individuen fehlen !
# Dieser eigentliche Kreuzungsprozess ist problemabhängig.
# Individuum x Individuum -> (list Individuum Individuum)
def kreuze(individuum1, individuum2):
    pos = waehlePositionen(len(individuum1))
    if pos[0]==0:
        neu1 = individuum2[:pos[1]]+individuum1[pos[1]:]
        neu2 = individuum1[:pos[1]]+individuum2[pos[1]:]
    elif pos[1]==len(individuum1)-1:
        neu1 = individuum2[:pos[0]]+individuum1[pos[0]:]
        neu2 = individuum1[:pos[0]]+individuum2[pos[0]:]
    else:
        neu1 = individuum1[:pos[0]]+individuum2[pos[0]:pos[1]]+individuum1[pos[1]:]
        neu2 = individuum2[:pos[0]]+individuum1[pos[0]:pos[1]]+individuum2[pos[1]:]
    return [neu1, neu2]

###Test:
##print(kreuze([15, 20, 92, 8, 41],[77,12,5,99,33]))

### ===== Kreuzen ===================================================
# Steuert das Kreuzen für die gesamte Population.
# Da beim Kreuzen neue Individuen erzeugt und der Population hinzu-
# gefügt werden, ändert sich die Gesamtzahl. Deren Anpassung erfolgt
# bei der Selektion.
# Die zu kreuzenden Individuen werden nach dem Zufallsprinzip ausge-
# wählt.
# Parameter sind: Population, crossing-over-Wahrscheinlichkeit
# Wert:  Population zuzüglich der Gekreuzten
def kreuzen(population, pCrossingOver):
    gekreuzte = []
    n=len(population)
    for individuum in population:
        if aktionP(pCrossingOver):
            anderes = population[random.randint(0,n-1)]
            while anderes == individuum:
                anderes = population[random.randint(0,n-1)]
            gekreuzte+=kreuze(individuum,anderes)
    return population+gekreuzte

###Test:
##population = erzeugePopulation()
##print(population)
##print()
##population = kreuzen(population, 40)
##for i in range(len(population)):
##    print(population[i])

### ===== besserP ===================================================
# Das Prädikat 'besserP' ist das Vergleichsprädikat zur Bewertung und
# ist daher vom speziellen Problem abhängig. Hier wird die vorgegebene
# Funktion bewerte (s.o.) genommen.
def besserP(individuum1, individuum2):
   return bewerte (individuum1) > bewerte (individuum2)

###Test:
##print(besserP([15, 20, 92, 8, 41],[80, 20, 93, 10, 1]))
##print(besserP([35, 89, 29, 88, 41],[80, 20, 93, 10, 1]))

### ===== selektiere ================================================
# Selektiert von zwei Individuen das mit der höheren Bewertungszahl.
def selektiere(individuum1, individuum2):
    if besserP(individuum1, individuum2):
        return individuum1
    return individuum2

###Test:
##print(selektiere([15, 20, 92, 8, 41],[80, 20, 93, 10, 1]))
##print(selektiere([35, 89, 29, 88, 41],[80, 20, 93, 10, 1]))

### ===== Selektion =================================================
# Es gibt viele veschiedene Selektionsmechanismen. Beispielsweise
# kann die Auswahl proportional zur Bewertungsfunktion erfolgen.
# Hier werden die vorhandenen Individuen einfach zufällig ausgewählt
# und einer paarweisen Auswahl unterzogen: Selektiert wird aus der
# Menge der Individuen durch direkten Vergleich von je zwei
# Individuen. Der Parameter Gesamtzahl gewährleistet, dass wieder
# die ursprüngliche Populationsgrösse hergestellt wird.
# Dazu wird mit der Selektion abgebrochen, wenn eine genügend grosse
# Zahl (-> Gesamtzahl) ausgewählt wurde.
def selektion(population, gesamtzahl):
    akku = []
    while gesamtzahl > len(akku):
        pos1,pos2 = waehlePositionen(len(population))
        akku.append(selektiere(population[pos1],population[pos2]))
    return akku

###Test:
##population = erzeugePopulation(20)
##print(population)
##print()
##print(selektion(population, 10))

### ===== Bewertung =================================================
# Führt die Bewertung der gesamten Population durch. Dazu wird die 
# Populationsliste in eine Liste mit Paaren [Individuum Bewertung]
# verwandelt.
def bewertung(population):
    akku = []
    for i in range(len(population)):
        akku.append([population[i],bewerte(population[i])])
    return akku

#####Test:
##population = erzeugePopulation(20)
##print(bewertung(population))

### ===== bestes-Individuum =========================================
# Für die grafische Darstellung, aber auch für das Ergebnis des GA
# ist es sinnvoll, das 'beste' Individuum bestimmen zu können.
def bestesIndividuum(population):
    indexBestes = 0
    for i in range(1,len(population)):
        if besserP(population[i],population[indexBestes]):
            indexBestes = i
    return population[indexBestes]

###Test:
##population = erzeugePopulation(20)
##for i in range(len(population)):
##    print(population[i]," --> ",bewerte(population[i]))
##print()
##print(bestesIndividuum(population))

### ===== Darstellung ===============================================
# Die Darstellung aller, also auch der 'besten' Lösung ist
# problemabhängig.
# Im Beispiel
def darstellung(individuum):
    print(individuum," --> ",bewerte(individuum))

### ===== genetischer-Algorithmus ===================================
def genetischerAlgorithmus(populationsGroesse,
                           anzahlSchritte,
                           mutationsWahrscheinlichkeit,
                           crossingOverWahrscheinlichkeit):
    population = erzeugePopulation(populationsGroesse)
    print(population)
    darstellung(bestesIndividuum(population))
    for i in range(anzahlSchritte):
        population= selektion(kreuzen(mutation(population,
                                               mutationsWahrscheinlichkeit),
                                      crossingOverWahrscheinlichkeit),
                              populationsGroesse)
        darstellung(bestesIndividuum(population))
    return population

###Test:
##population = genetischerAlgorithmus(10,10,30,40)
##for i in range(len(population)):
##    print(population[i]," --> ",bewerte(population[i]))
##print()
##print(bestesIndividuum(population))

### ===== GA-Aufruf =================================================
# Eine Aufrufhülle für 'genetischerAlgorithmus', bei der die
# Parameter zur Laufzeit eingegeben werden können
def GA_Aufruf():
    populationsGroesse=int(input("Gib die Zahl der Individuen an [> 0] : "))
    anzahlSchritte=int(input("Wieviel Schritte ?             [> 0] : "))
    mutationsWahrscheinlichkeit=int(input("Mutationswahrscheinlichkeit ?  [00 .. 99] : "))
    crossingOverWahrscheinlichkeit=int(input("Wahrscheinlichkeit für crossing-over ? [00 .. 99] : "))
    genetischerAlgorithmus(populationsGroesse,
                           anzahlSchritte,
                           mutationsWahrscheinlichkeit,
                           crossingOverWahrscheinlichkeit)

### ===== Aufruf ====================================================
genetischerAlgorithmus(20, 20, 40, 40)
